from typing import TypeVar, Type, Generator

from cfinterface.components.register import Register


class RegisterData:
    """
    Class for a storing, managing and accessing data for a register file.
    """

    T = TypeVar("T")

    def __init__(self, root: Register) -> None:
        self.__root = root
        self.__head = root

    def __iter__(self):
        current = self.__root
        while current:
            yield current
            current = current.next

    def __len__(self) -> int:
        count = 0
        for _ in self:
            count += 1
        return count

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, RegisterData):
            return False
        rd: RegisterData = o
        if len(self) != len(rd):
            return False
        for r1, r2 in zip(self, rd):
            if r1 != r2:
                return False
        return True

    def preppend(self, r: Register):
        """
        Appends a register to the beginning of the data.

        :param r: The new register to preppended to the data.
        :type r: Register
        """
        self.add_before(self.__root, r)

    def append(self, r: Register):
        """
        Appends a register to the end of the data.

        :param r: The new register to append to the data
        :type r: Register
        """
        self.add_after(self.__head, r)

    def add_before(self, before: Register, new: Register):
        """
        Adds a new register to the data before another
        specified register.

        :param before: The existing register which will be preppended
        :type before: Register
        :param new: The new register to add to the data
        :type new: Register
        """
        if before == self.__root:
            self.__root = new
        new.previous = before.previous
        before.previous = new
        new.next = before

    def add_after(self, after: Register, new: Register):
        """
        Adds a new register to the data after another
        specified register.

        :param after: The existing register which will be appended
        :type after: Register
        :param new: The new register to add to the data
        :type new: Register
        """
        if after == self.__head:
            self.__head = new
        new.next = after.next
        after.next = new
        new.previous = after

    def remove(self, r: Register):
        """
        Removes an existing register in the chain.

        :param r: The register to be removed
        :type r: Register
        """
        if r.previous is not None:
            r.previous.next = r.next
        if r.next is not None:
            r.next.previous = r.previous

    def of_type(self, t: Type[T]) -> Generator[T, None, None]:
        """
        A register generator that only returns registers of type T.

        :param t: The register type that is desired
        :type t: Type[T]
        :yield: Registers filtered by type T
        :rtype: Generator[T, None, None]
        """
        for r in self:
            if isinstance(r, t):
                yield r

    @property
    def first(self) -> Register:
        return self.__root

    @property
    def last(self) -> Register:
        return self.__head



from typing import Any, List, Optional

from cfinterface.components.field import Field

from cfinterface.adapters.repository import Repository
from cfinterface.adapters.positionalrepository import PositionalRepository
from cfinterface.adapters.delimitedrepository import DelimitedRepository


class Line:
    """
    Class for representing a generic line that is composed
    of fields and can be read from or written to a file.
    """

    def __init__(
        self,
        fields: List[Field],
        values: Optional[List[Any]] = None,
        delimiter: Optional[str] = None,
    ):
        self._delimiter = delimiter
        self._repository: Repository = None  # type: ignore
        if delimiter is None:
            self._repository = PositionalRepository(fields, values)
        else:
            self._repository = DelimitedRepository(fields, values, delimiter)

    def read(self, line: str) -> List[Any]:
        """
        Reads a line for extracting information following
        the given fields.

        :param line: The line to be read
        :type line: str
        :return: The extracted values, in order
        :rtype: List[Any]
        """
        return self._repository.read(line)

    def write(self, values: List[Any]) -> str:
        """
        Writes to a line with the existing information
        in the given fields.

        :param values: The value of the fields to be written
        :type line: List[Any]
        :return: The line with the new field information
        :rtype: str
        """
        return self._repository.write(values)

    @property
    def fields(self) -> List[Field]:
        return self._repository.fields

    @fields.setter
    def fields(self, vals: List[Field]):
        self._repository.fields = vals

    @property
    def values(self) -> List[Any]:
        return self._repository.values

    @values.setter
    def values(self, vals: List[Any]):
        self._repository.values = vals

    @property
    def delimiter(self) -> Optional[str]:
        return self._delimiter

    @delimiter.setter
    def delimiter(self, d: Optional[str]):
        self._delimiter = d



from cfinterface.components.defaultregister import DefaultRegister
from cfinterface.components.state import ComponentState
from tests.mocks.mock_open import mock_open

from unittest.mock import MagicMock, patch


def test_default_register_eq():
    r1 = DefaultRegister(data=[0])
    r2 = DefaultRegister(data=[0])
    assert r1 == r2
    r1.data = [1]
    assert r1 != r2


def test_default_register_read():
    data = "Hello, world!\n"
    m: MagicMock = mock_open(read_data=data)
    with patch("builtins.open", m):
        with open("", "r") as fp:
            r = DefaultRegister()
            r.read_register(fp)
            assert r.data == data
            assert r.success


def test_default_register_write():
    data = "Hello, world!\n"
    filedata = ""
    m = mock_open(read_data=filedata)
    with patch("builtins.open", m):
        with open("", "w") as fp:
            r = DefaultRegister(state=ComponentState.READ_SUCCESS)
            r.data = data
            r.write_register(fp)
    m().write.assert_called_once_with(data)



from typing import IO, List

from cfinterface.components.section import Section
from cfinterface.data.sectiondata import SectionData
from cfinterface.writing.sectionwriting import SectionWriting

from tests.mocks.mock_open import mock_open

from unittest.mock import MagicMock, patch


class DummySection(Section):
    def __eq__(self, o: object) -> bool:
        if not isinstance(o, self.__class__):
            return False
        else:
            return o.data == self.data

    def read(self, file: IO) -> bool:
        self.data: List[str] = []
        line: str = file.readline()
        self.data.append(line)
        return True

    def write(self, file: IO) -> bool:
        file.write(self.data)
        return True


def test_sectionwriting_withdata():
    filedata = "Hello, World!"
    bd = SectionData(DummySection(data=filedata))
    bw = SectionWriting(bd)
    m: MagicMock = mock_open(read_data=filedata)
    with patch("builtins.open", m):
        bw.write("", "", "utf-8")
    m().write.assert_called_once_with(filedata)



from cfinterface.components.floatfield import FloatField


def test_floatfield_read():
    data = 12345
    field = FloatField(5, 0, 1)
    line = f"{data}-something-else"
    field.read(line)
    assert field.value == data


def test_floatfield_read_scientific_notation():
    data = "1.2e3"
    field = FloatField(5, 0, 1, format="e")
    line = f"{data}-something-else"
    field.read(line)
    assert field.value == float(data)


def test_floatfield_comma_separator():
    data = "1,23"
    field = FloatField(4, 0, 1, sep=",")
    line = f"{data}-something-else"
    field.read(line)
    assert field.value == 1.23


def test_floatfield_write():
    line_before = "field-12345-else"
    data = float(line_before[6:11])
    field = FloatField(5, 6, 1, value=data)
    line_after = field.write(line_before)
    assert line_before == line_after


def test_floatfield_write_empty():
    field = FloatField(5, 0, 1)
    assert len(field.write("")) == 5


def test_floatfield_write_short_line():
    line_before = "field-12345-else"
    data = float(line_before[6:11])
    field = FloatField(5, 6, 1, value=data)
    line_after = field.write("   ")
    assert data == float(line_after[6:])



from typing import List, Any

from cfinterface.adapters.repository import Repository


class PositionalRepository(Repository):

    # Override
    def read(self, line: str) -> List[Any]:
        """
        Reads a line for extracting information following
        the given fields.

        :param line: The line to be read
        :type line: str
        :return: The extracted values, in order
        :rtype: List[Any]
        """
        for field in self._fields:
            field.read(line)
        return self.values

    # Override
    def write(self, values: List[Any]) -> str:
        """
        Writes to a line with the existing information
        in the given fields.

        :param values: The value of the fields to be written
        :type line: List[Any]
        :return: The line with the new field information
        :rtype: str
        """
        line = ""
        self.values = values
        for field in self._fields:
            line = field.write(line)
        return line + "\n"



from cfinterface.components.line import Line
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.register import Register
from cfinterface.reading.registerreading import RegisterReading

from tests.mocks.mock_open import mock_open

from unittest.mock import MagicMock, patch


class DummyRegister(Register):
    IDENTIFIER = "reg"
    IDENTIFIER_DIGITS = 4
    LINE = Line([LiteralField(13, 4)])


def test_registerreading_empty():
    filedata = ""
    br = RegisterReading([DummyRegister])
    m: MagicMock = mock_open(read_data=filedata)
    with patch("builtins.open", m):
        bd = br.read("", "", "utf-8")
        assert br.empty
        assert len(bd) == 1


def test_registerreading_withdata():
    data = "Hello, world!"
    br = RegisterReading([DummyRegister])
    m: MagicMock = mock_open(
        read_data=DummyRegister.IDENTIFIER + " " + data + "\n"
    )
    with patch("builtins.open", m):
        bd = br.read("", "", "utf-8")
        assert not br.empty
        dbs = [b for b in bd.of_type(DummyRegister)]
        assert len(dbs) == 1
        assert dbs[0].data[0].strip() == data



from typing import List, Any, Optional

from cfinterface.components.field import Field
from cfinterface.adapters.repository import Repository


class DelimitedRepository(Repository):
    def __init__(
        self,
        fields: List[Field],
        values: Optional[List[Any]] = None,
        delimiter: str = ";",
    ) -> None:
        self._delimter = delimiter
        self._fields = [
            DelimitedRepository.__positional_to_delimited_field(f)
            for f in fields
        ]
        if values is not None:
            for f, v in zip(self._fields, values):
                f.value = v

    @staticmethod
    def __positional_to_delimited_field(f: Field) -> Field:
        f.ending_column = f.size
        f.starting_column = 0
        return f

    # Override
    def read(self, line: str) -> List[Any]:
        """
        Reads a line for extracting information following
        the given fields.

        :param line: The line to be read
        :type line: str
        :return: The extracted values, in order
        :rtype: List[Any]
        """
        values = [v.strip() for v in line.split(self._delimter)]
        for field, value in zip(self._fields, values):
            field.read(value)
        return self.values

    # Override
    def write(self, values: List[Any]) -> str:
        """
        Writes to a line with the existing information
        in the given fields.

        :param values: The value of the fields to be written
        :type line: List[Any]
        :return: The line with the new field information
        :rtype: str
        """
        self.values = values
        separated = [field.write("") for field in self._fields]
        return self._delimter.join(separated) + "\n"



from cfinterface.components.section import Section
from cfinterface.data.sectiondata import SectionData


class DummySection(Section):
    def __eq__(self, o: object) -> bool:
        if not isinstance(o, self.__class__):
            return False
        else:
            return o.data == self.data


def test_blockdata_eq():
    sd1 = SectionData(DummySection(data=-1))
    sd2 = SectionData(DummySection(data=-1))
    assert sd1 == sd2


def test_blockdata_not_eq():
    sd1 = SectionData(DummySection(data=-1))
    sd2 = SectionData(DummySection(data=+1))
    assert sd1 != sd2


def test_blockdata_append():
    sd = SectionData(DummySection(data=-1))
    n_sections = 10
    for i in range(n_sections):
        sd.append(DummySection(data=i))
    assert len(sd) == n_sections + 1
    assert sd.first.data == -1
    assert sd.last.data == n_sections - 1


def test_blockdata_preppend():
    sd = SectionData(DummySection(data=-1))
    n_sections = 10
    for i in range(n_sections):
        sd.preppend(DummySection(data=i))
    assert len(sd) == n_sections + 1
    assert sd.first.data == n_sections - 1
    assert sd.last.data == -1


def test_blockdata_add_before():
    sd = SectionData(DummySection(data=-1))
    s1 = DummySection(data=1)
    sd.append(s1)
    s2 = DummySection(data=2)
    sd.add_before(s1, s2)
    assert s1.previous == s2
    assert s2.next == s1


def test_blockdata_add_after():
    sd = SectionData(DummySection(data=-1))
    s1 = DummySection(data=1)
    sd.append(s1)
    s2 = DummySection(data=2)
    sd.add_after(s1, s2)
    assert s1.next == s2
    assert s2.previous == s1


def test_blockdata_remove():
    sd = SectionData(DummySection(data=-1))
    s1 = DummySection(data=1)
    sd.append(s1)
    assert len(sd) == 2
    sd.remove(s1)
    assert len(sd) == 1


def test_blockdata_of_type():
    sd = SectionData(DummySection())
    sd.append(Section())
    assert len(sd) == 2
    assert len([b for b in sd.of_type(DummySection)]) == 1



from typing import IO, List

from cfinterface.components.section import Section
from cfinterface.reading.sectionreading import SectionReading

from tests.mocks.mock_open import mock_open

from unittest.mock import MagicMock, patch


class DummySection(Section):
    def __eq__(self, o: object) -> bool:
        if not isinstance(o, self.__class__):
            return False
        else:
            return o.data == self.data

    def read(self, file: IO) -> bool:
        self.data: List[str] = []
        self.data.append(file.readline())
        return True

    def write(self, file: IO) -> bool:
        for line in self.data:
            file.write(line)
        return True


def test_sectionreading_withdata():
    data = "Hello, world!"
    sr = SectionReading([DummySection])
    m: MagicMock = mock_open(read_data=data + "\n")
    with patch("builtins.open", m):
        sd = sr.read("", "", "utf-8")
        assert not sr.empty
        dbs = [b for b in sd.of_type(DummySection)]
        assert len(dbs) == 1
        assert dbs[0].data[0].strip() == data



from typing import List, Type

from cfinterface.components.register import Register
from cfinterface.components.defaultregister import DefaultRegister
from cfinterface.data.registerdata import RegisterData
from cfinterface.reading.registerreading import RegisterReading
from cfinterface.writing.registerwriting import RegisterWriting


class RegisterFile:
    """
    Class that models a file divided by registers, where the reading
    and writing are given by a series of registers.
    """

    REGISTERS: List[Type[Register]] = []
    ENCODING = "utf-8"

    def __init__(
        self,
        data=RegisterData(DefaultRegister("")),
    ) -> None:
        self.__data = data

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, RegisterFile):
            return False
        bf: RegisterFile = o
        return self.data == bf.data

    @classmethod
    def read(cls, directory: str, filename: str = ""):
        """
        Reads the registerfile data from a given file in disk.

        :param filename: The file name in disk
        :type filename: str
        :param directory: The directory where the file is
        :type directory: str
        """
        reader = RegisterReading(cls.REGISTERS)
        return cls(reader.read(filename, directory, cls.ENCODING))

    def write(self, directory: str, filename: str = ""):
        """
        Write the registerfile data to a given file in disk.

        :param filename: The file name in disk
        :type filename: str
        :param directory: The directory where the file will be
        :type directory: str
        """
        writer = RegisterWriting(self.__data)
        writer.write(filename, directory, self.__class__.ENCODING)

    @property
    def data(self) -> RegisterData:
        return self.__data



from typing import Any, IO
import re
from cfinterface.components.field import Field

from cfinterface.components.literalfield import LiteralField
from cfinterface.components.state import ComponentState
from cfinterface.components.line import Line


class Register:
    """
    Class for a generic register in a textfile, with a given identifier
    which is located at the beginning of the line.
    """

    IDENTIFIER = ""
    IDENTIFIER_DIGITS = 0
    LINE = Line([])

    def __init__(
        self,
        state=ComponentState.NOT_FOUND,
        previous=None,
        next=None,
        data=None,
    ) -> None:
        identifier_field: Field = LiteralField(
            self.__class__.IDENTIFIER_DIGITS, 0
        )
        self.__line = Line(
            [identifier_field] + self.__class__.LINE.fields,
            delimiter=self.__class__.LINE.delimiter,
        )
        self.__state = state
        self.__previous = previous
        self.__next = next
        self.__data: Any = data

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, self.__class__):
            return False
        else:
            return o.data == self.data

    @classmethod
    def matches(cls, line: str):
        """
        Checks if the current line matches the identifier of the register.

        :param line: The candidate line for containing
            the register information
        :type line: str
        """
        return (
            re.search(cls.IDENTIFIER, line[: cls.IDENTIFIER_DIGITS])
            is not None
        )

    def read(self, file: IO) -> bool:
        """
        Generic function to perform the reading of the register using
        a filepointer.

        :param file: The filepointer
        :type file: IO
        :return: The success, or not, in the reading
        :rtype: bool
        """
        self.data = self.__line.read(file.readline())[1:]
        return True

    def write(self, file: IO) -> bool:
        """
        Generic function to perform the writing of the register using
        a filepointer.

        :param file: The filepointer
        :type file: IO
        :return: The success, or not, in the writing
        :rtype: bool
        """
        line = self.__line.write([self.__class__.IDENTIFIER] + self.data)
        file.write(line)
        return True

    def read_register(self, file: IO):
        """
        Function that reads the register and evaluates the result.

        :param file: The filepointer
        :type file: IO
        """
        if self.read(file):
            self.__state = ComponentState.READ_SUCCESS
        else:
            self.__state = ComponentState.READ_ERROR

    def write_register(self, file: IO):
        """
        Function that writes the register, if it was succesfully read.

        :param file: The filepointer
        :type file: IO
        """
        if self.__state == ComponentState.READ_SUCCESS:
            if self.write(file):
                self.__state = ComponentState.WRITE_SUCCESS
            else:
                self.__state = ComponentState.WRITE_ERROR

    @property
    def previous(self) -> "Register":
        return self.__previous

    @previous.setter
    def previous(self, b: "Register"):
        self.__previous = b

    @property
    def next(self) -> "Register":
        return self.__next

    @next.setter
    def next(self, b: "Register"):
        self.__next = b

    @property
    def data(self) -> Any:
        return self.__data

    @data.setter
    def data(self, d: Any):
        self.__data = d

    @property
    def is_first(self) -> bool:
        return self.__previous is None

    @property
    def is_last(self) -> bool:
        return self.__next is None

    @property
    def empty(self) -> bool:
        return self.__data is None

    @property
    def success(self) -> bool:
        return self.__state in [
            ComponentState.READ_SUCCESS,
            ComponentState.WRITE_SUCCESS,
        ]



"""
cfi
====

cfi is a Python module for handling custom formatted files
and provide reading, storing and writing utilities.
"""

__version__ = "0.0.14"

from . import components  # noqa
from . import data  # noqa
from . import reading  # noqa
from . import writing  # noqa
from . import files  # noqa



from typing import IO
from cfinterface.components.section import Section


class DefaultSection(Section):
    """
    A class for representing a default section, which contains no data
    and is used for representing empty data.
    """

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, DefaultSection):
            return False
        return self.data == o.data

    def read(self, file: IO) -> bool:
        self.data = file.readline()
        return True

    def write(self, file: IO) -> bool:
        file.write(self.data)
        return True



from typing import IO, List

from cfinterface.components.line import Line
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.register import Register
from cfinterface.data.registerdata import RegisterData
from cfinterface.writing.registerwriting import RegisterWriting

from tests.mocks.mock_open import mock_open

from unittest.mock import MagicMock, patch


class DummyRegister(Register):
    IDENTIFIER = "reg"
    IDENTIFIER_DIGITS = 4
    LINE = Line([LiteralField(13, 4)])


def test_blockwriting_withdata():
    filedata = "Hello, World!"
    bd = RegisterData(DummyRegister(data=[filedata]))
    bw = RegisterWriting(bd)
    m: MagicMock = mock_open(read_data=filedata)
    with patch("builtins.open", m):
        bw.write("", "", "utf-8")
    m().write.assert_called_once_with(
        DummyRegister.IDENTIFIER + " " + filedata + "\n"
    )



from cfinterface.components.line import Line
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.register import Register
from cfinterface.data.registerdata import RegisterData


class DummyRegister(Register):
    IDENTIFIER = "reg"
    IDENTIFIER_DIGITS = 4
    LINE = Line([LiteralField(13, 4)])


def test_registerdata_eq():
    rd1 = RegisterData(DummyRegister(data=-1))
    rd2 = RegisterData(DummyRegister(data=-1))
    assert rd1 == rd2


def test_registerdata_not_eq():
    rd1 = RegisterData(DummyRegister(data=-1))
    rd2 = RegisterData(DummyRegister(data=+1))
    assert rd1 != rd2


def test_registerdata_append():
    rd = RegisterData(DummyRegister(data=-1))
    n_registers = 10
    for i in range(n_registers):
        rd.append(DummyRegister(data=i))
    assert len(rd) == n_registers + 1
    assert rd.first.data == -1
    assert rd.last.data == n_registers - 1


def test_registerdata_preppend():
    rd = RegisterData(DummyRegister(data=-1))
    n_registers = 10
    for i in range(n_registers):
        rd.preppend(DummyRegister(data=i))
    assert len(rd) == n_registers + 1
    assert rd.first.data == n_registers - 1
    assert rd.last.data == -1


def test_registerdata_add_before():
    rd = RegisterData(DummyRegister(data=-1))
    r1 = DummyRegister(data=1)
    rd.append(r1)
    r2 = DummyRegister(data=2)
    rd.add_before(r1, r2)
    assert r1.previous == r2
    assert r2.next == r1


def test_registerdata_add_after():
    rd = RegisterData(DummyRegister(data=-1))
    r1 = DummyRegister(data=1)
    rd.append(r1)
    r2 = DummyRegister(data=2)
    rd.add_after(r1, r2)
    assert r1.next == r2
    assert r2.previous == r1


def test_registerdata_remove():
    rd = RegisterData(DummyRegister(data=-1))
    r1 = DummyRegister(data=1)
    rd.append(r1)
    assert len(rd) == 2
    rd.remove(r1)
    assert len(rd) == 1


def test_registerdata_of_type():
    rd = RegisterData(DummyRegister())
    rd.append(Register())
    assert len(rd) == 2
    assert len([b for b in rd.of_type(DummyRegister)]) == 1



from typing import TypeVar, Type, Generator

from cfinterface.components.section import Section


class SectionData:
    """
    Class for a storing, managing and accessing data for a section file.
    """

    T = TypeVar("T")

    def __init__(self, root: Section) -> None:
        self.__root = root
        self.__head = root

    def __iter__(self):
        current = self.__root
        while current:
            yield current
            current = current.next

    def __len__(self) -> int:
        count = 0
        for _ in self:
            count += 1
        return count

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, SectionData):
            return False
        sd: SectionData = o
        if len(self) != len(sd):
            return False
        for r1, r2 in zip(self, sd):
            if r1 != r2:
                return False
        return True

    def preppend(self, s: Section):
        """
        Appends a section to the beginning of the data.

        :param s: The new section to preppended to the data.
        :type s: Section
        """
        self.add_before(self.__root, s)

    def append(self, s: Section):
        """
        Appends a section to the end of the data.

        :param s: The new section to append to the data
        :type s: Section
        """
        self.add_after(self.__head, s)

    def add_before(self, before: Section, new: Section):
        """
        Adds a new section to the data before another
        specified section.

        :param before: The existing section which will be preppended
        :type before: Section
        :param new: The new section to add to the data
        :type new: Section
        """
        if before == self.__root:
            self.__root = new
        new.previous = before.previous
        before.previous = new
        new.next = before

    def add_after(self, after: Section, new: Section):
        """
        Adds a new section to the data after another
        specified section.

        :param after: The existing section which will be appended
        :type after: Section
        :param new: The new section to add to the data
        :type new: Section
        """
        if after == self.__head:
            self.__head = new
        new.next = after.next
        after.next = new
        new.previous = after

    def remove(self, s: Section):
        """
        Removes an existing section in the chain.

        :param s: The section to be removed
        :type s: Section
        """
        if s.previous is not None:
            s.previous.next = s.next
        if s.next is not None:
            s.next.previous = s.previous

    def of_type(self, t: Type[T]) -> Generator[T, None, None]:
        """
        A section generator that only returns sections of type T.

        :param t: The section type that is desired
        :type t: Type[T]
        :yield: Sections filtered by type T
        :rtype: Generator[T, None, None]
        """
        for s in self:
            if isinstance(s, t):
                yield s

    @property
    def first(self) -> Section:
        return self.__root

    @property
    def last(self) -> Section:
        return self.__head



from cfinterface.adapters.delimitedrepository import DelimitedRepository
from cfinterface.components.literalfield import LiteralField


def test_delimitedrepository_read_no_fields():
    repo = DelimitedRepository([])
    fileline = ""
    assert len(repo.read(fileline)) == 0


def test_delimitedrepository_read_with_fields():
    fields = [LiteralField(6, 0), LiteralField(6, 7)]
    repo = DelimitedRepository(fields)
    fileline = "hello,;world!"
    values = repo.read(fileline)
    assert values[0] == "hello,"
    assert values[1] == "world!"


def test_delimitedrepository_write_no_fields():
    repo = DelimitedRepository([])
    assert len(repo.write([])) == 1


def test_delimitedrepository_write_with_fields():
    fields = [LiteralField(6, 0), LiteralField(6, 7)]
    values = ["hello,", "world!"]
    repo = DelimitedRepository(fields, values)
    fileline = "hello,;world!\n"
    outline = repo.write(values)
    assert fileline == outline



from cfinterface.components.line import Line
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.register import Register
from cfinterface.components.state import ComponentState
from cfinterface.data.registerdata import RegisterData
from cfinterface.files.registerfile import RegisterFile

from tests.mocks.mock_open import mock_open

from unittest.mock import MagicMock, patch


class DummyRegister(Register):
    IDENTIFIER = "reg"
    IDENTIFIER_DIGITS = 4
    LINE = Line([LiteralField(13, 4)])


def test_registerfile_eq():
    rf1 = RegisterFile(data=RegisterData(DummyRegister(data=-1)))
    rf2 = RegisterFile(data=RegisterData(DummyRegister(data=-1)))
    assert rf1 == rf2


def test_registerfile_not_eq_invalid_type():
    rf1 = RegisterFile(data=RegisterData(DummyRegister(data=-1)))
    rf2 = 5
    assert rf1 != rf2


def test_registerfile_not_eq_different_length():
    bd = RegisterData(DummyRegister(data=-1))
    bd.append(DummyRegister(data=+1))
    rf1 = RegisterFile(data=bd)
    rf2 = RegisterFile(data=RegisterData(DummyRegister(data=-1)))
    assert rf1 != rf2


def test_registerfile_not_eq_valid():
    rf1 = RegisterFile(data=RegisterData(DummyRegister(data=-1)))
    rf2 = RegisterFile(data=RegisterData(DummyRegister(data=+1)))
    assert rf1 != rf2


def test_registerfile_read():
    data = "Hello, world!"
    filedata = DummyRegister.IDENTIFIER + " " + data + "\n"
    RegisterFile.REGISTERS = [DummyRegister]
    m: MagicMock = mock_open(read_data=filedata)
    with patch("builtins.open", m):
        f = RegisterFile.read("", "")
        assert len(f.data) == 2
        assert f.data.last.data[0] == data


def test_registerfile_write():
    data = "Hello, world!"
    bd = RegisterData(
        DummyRegister(state=ComponentState.READ_SUCCESS, data=[data])
    )
    RegisterFile.REGISTERS = [DummyRegister]
    f = RegisterFile(bd)
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        f.write("", "")
    m().write.assert_called_once_with(
        DummyRegister.IDENTIFIER + " " + data + "\n"
    )



from typing import IO

import pytest

from cfinterface.components.section import Section
from cfinterface.components.state import ComponentState
from tests.mocks.mock_open import mock_open

from unittest.mock import MagicMock, patch


class DummySection(Section):
    def __eq__(self, o: object) -> bool:
        if not isinstance(o, self.__class__):
            return False
        else:
            return o.data == self.data

    def read(self, file: IO) -> bool:
        self.data = file.readline().strip()
        return True

    def write(self, file: IO) -> bool:
        file.write(self.data)
        return True


def test_single_section_success():
    s1 = Section(state=ComponentState.READ_SUCCESS)
    assert s1.success


def test_single_section_not_found_error():
    s1 = Section(state=ComponentState.NOT_FOUND)
    assert not s1.success


def test_single_section_read_error():
    s1 = Section(state=ComponentState.READ_ERROR)
    assert not s1.success


def test_single_section_properties():
    s1 = Section()
    assert s1.is_first
    assert s1.is_last


def test_section_simple_chain_properties():
    # Build a simple section chain
    s1 = Section()
    s2 = Section()
    s3 = Section()
    # Sets relationships
    s1.next = s2
    s2.previous = s1
    s2.next = s3
    s3.previous = s2
    # Asserts properties
    assert s1.is_first
    assert s3.is_last
    assert not s1.is_last
    assert not s2.is_first
    assert not s2.is_last
    assert not s3.is_first
    assert s1.empty
    assert s2.empty
    assert s3.empty


def test_section_equal_error():
    s1 = Section()
    s2 = Section()
    with pytest.raises(NotImplementedError):
        s1 == s2


def test_section_read_error():
    s = Section()
    with pytest.raises(NotImplementedError):
        m: MagicMock = mock_open(read_data="")
        with patch("builtins.open", m):
            with open("", "r") as fp:
                s.read_section(fp)


def test_section_write_error():
    s = Section(state=ComponentState.READ_SUCCESS)
    with pytest.raises(NotImplementedError):
        m: MagicMock = mock_open(read_data="")
        with patch("builtins.open", m):
            with open("", "r") as fp:
                s.write_section(fp)


def test_dummy_section_equal():
    s1 = DummySection()
    s2 = DummySection()
    assert s1 == s2


def test_dummy_section_read():
    data = "Hello, world!"
    m: MagicMock = mock_open(read_data=data)
    with patch("builtins.open", m):
        with open("", "r") as fp:
            s = DummySection()
            s.read_section(fp)
            assert s.data == data
            assert s.success


def test_dummy_block_write():
    data = "Hello, world!"
    filedata = ""
    m = mock_open(read_data=filedata)
    with patch("builtins.open", m):
        with open("", "w") as fp:
            b = DummySection(state=ComponentState.READ_SUCCESS)
            b.data = data
            b.write_section(fp)
    m().write.assert_called_once_with(data)



from typing import IO, List, Type
from os.path import join

from cfinterface.components.section import Section
from cfinterface.components.defaultsection import DefaultSection
from cfinterface.data.sectiondata import SectionData


class SectionReading:
    """
    Class for reading custom files based on a SectionData structure.
    """

    def __init__(self, sections: List[Type[Section]]) -> None:
        self.__sections = sections
        self.__data = SectionData(DefaultSection(data=""))
        self.__last_position_filepointer = 0

    def __read_line_with_backup(self, file: IO) -> str:
        """
        Reads a line of the file, saving the filepointer position
        in case one desired to return to the previous line.

        :param file: THe filepoiner for the reading file
        :type file: IO
        :return: The read line
        :rtype: str
        """
        self.__last_position_filepointer = file.tell()
        return file.readline()

    def __restore_previous_line(self, file: IO):
        """
        Restores the filepointer to the beginning of the previously
        read line.

        :param file: The filepointer
        :type file: IO
        """
        file.seek(self.__last_position_filepointer)

    def __read_file(self, file: IO) -> SectionData:
        """
        Reads all the sections from a file and
        returns the SectionData structure.

        :param file: The filepointer
        :type file: IO
        :return: The section data from the file
        :rtype: SectionData
        """
        for sectiontype in self.__sections:
            section = sectiontype()
            section.read(file)
            self.__data.append(section)
        while True:
            line = self.__read_line_with_backup(file)
            if len(line) == 0:
                break
            self.__restore_previous_line(file)
            section = DefaultSection()
            section.read(file)
            self.__data.append(section)
        return self.__data

    def read(
        self, filename: str, directory: str, encoding: str
    ) -> SectionData:
        """
        Reads a file with a given name in a given directory and
        extracts the data from the specified sections.

        :param filename: The name of the file
        :type filename: str
        :param directory: The directory where the file is
        :type directory: str
        :param encoding: The encoding for reading the file
        :type encoding: str
        :return: The data from the sections found in the file
        :rtype: SectionData
        """
        filepath = join(directory, filename)
        with open(filepath, "r", encoding=encoding) as fp:
            return self.__read_file(fp)

    @property
    def data(self) -> SectionData:
        return self.__data

    @property
    def empty(self) -> bool:
        return len(self.__data) == 1



from cfinterface.components.line import Line
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.register import Register
from cfinterface.components.state import ComponentState
from tests.mocks.mock_open import mock_open

from unittest.mock import MagicMock, patch


class DummyRegister(Register):
    IDENTIFIER = "reg"
    IDENTIFIER_DIGITS = 4
    LINE = Line([LiteralField(13, 4)])


class DummyDelimitedRegister(Register):
    IDENTIFIER = "reg"
    IDENTIFIER_DIGITS = 4
    LINE = Line([LiteralField(13, 4)], delimiter=";")


def test_single_register_success():
    r1 = Register(state=ComponentState.READ_SUCCESS)
    assert r1.success


def test_single_register_not_found_error():
    r1 = Register(state=ComponentState.NOT_FOUND)
    assert not r1.success


def test_single_register_read_error():
    r1 = Register(state=ComponentState.READ_ERROR)
    assert not r1.success


def test_single_register_properties():
    r1 = Register()
    assert r1.is_first
    assert r1.is_last


def test_register_simple_chain_properties():
    # Build a simple register chain
    r1 = Register()
    r2 = Register()
    r3 = Register()
    # Sets relationships
    r1.next = r2
    r2.previous = r1
    r2.next = r3
    r3.previous = r2
    # Asserts properties
    assert r1.is_first
    assert r3.is_last
    assert not r1.is_last
    assert not r2.is_first
    assert not r2.is_last
    assert not r3.is_first
    assert r1.empty
    assert r2.empty
    assert r3.empty


def test_dummy_register_equal():
    b1 = DummyRegister()
    b2 = DummyRegister()
    assert b1 == b2


def test_dummy_register_read():
    data = "Hello, world!"
    filedata = DummyRegister.IDENTIFIER + " " + data + "\n"
    m: MagicMock = mock_open(read_data=filedata)
    with patch("builtins.open", m):
        with open("", "r") as fp:
            b = DummyRegister()
            b.read_register(fp)
            assert b.data[0] == data
            assert b.success


def test_dummy_register_write():
    data = "Hello, world!"
    write_data = DummyRegister.IDENTIFIER + " " + data + "\n"
    filedata = ""
    m = mock_open(read_data=filedata)
    with patch("builtins.open", m):
        with open("", "w") as fp:
            b = DummyRegister(state=ComponentState.READ_SUCCESS)
            b.data = [data]
            b.write_register(fp)
    m().write.assert_called_once_with(write_data)


def test_dummy_delimiterregister_read():
    data = "Hello, world!"
    filedata = DummyDelimitedRegister.IDENTIFIER + " ;" + data + "\n"
    m: MagicMock = mock_open(read_data=filedata)
    with patch("builtins.open", m):
        with open("", "r") as fp:
            b = DummyDelimitedRegister()
            b.read_register(fp)
            assert b.data[0] == data
            assert b.success


def test_dummy_delimiterregister_write():
    data = "Hello, world!"
    write_data = DummyDelimitedRegister.IDENTIFIER + " ;" + data + "\n"
    filedata = ""
    m = mock_open(read_data=filedata)
    with patch("builtins.open", m):
        with open("", "w") as fp:
            b = DummyDelimitedRegister(state=ComponentState.READ_SUCCESS)
            b.data = [data]
            b.write_register(fp)
    m().write.assert_called_once_with(write_data)



from typing import List, Optional, Any
from abc import ABC, abstractmethod

from cfinterface.components.field import Field


class Repository(ABC):
    def __init__(
        self, fields: List[Field], values: Optional[List[Any]] = None
    ) -> None:
        self._fields = fields
        if values is not None:
            for f, v in zip(self._fields, values):
                f.value = v

    @abstractmethod
    def read(self, line: Any) -> List[Any]:
        """
        Reads a line for extracting information following
        the given fields.

        :param line: The line to be read
        :type line: Any
        :return: The extracted values, in order
        :rtype: List[Any]
        """
        raise NotImplementedError

    @abstractmethod
    def write(self, values: List[Any]) -> Any:
        """
        Writes to a line with the existing information
        in the given fields.

        :param values: The value of the fields to be written
        :type line: List[Any]
        :return: The line with the new field information
        :rtype: Any
        """
        raise NotImplementedError

    @property
    def fields(self) -> List[Field]:
        return self._fields

    @fields.setter
    def fields(self, f: List[Field]):
        self._fields = f

    @property
    def values(self) -> List[Any]:
        return [f.value for f in self._fields]

    @values.setter
    def values(self, vals: List[Any]):
        for f, v in zip(self._fields, vals):
            f.value = v
